-- VECTRIC LUA SCRIPT
-- Copyright 2014 Vectric Ltd.

-- Gadgets are an entirely optional add-in to Vectric's core software products. 
-- They are provided 'as-is', without any express or implied warranty, and you make use of them entirely at your own risk.
-- In no event will the author(s) or Vectric Ltd. be held liable for any damages arising from their use.

-- Permission is granted to anyone to use this software for any purpose, 
-- including commercial applications, and to alter it and redistribute it freely, 
-- subject to the following restrictions:

-- 1. The origin of this software must not be misrepresented; you must not claim that you wrote the original software. If you use this software in a product, an acknowledgment in the product documentation would be appreciated but is not required.
-- 2. Altered source versions must be plainly marked as such, and must not be misrepresented as being the original software.
-- 3. This notice may not be removed or altered from any source distribution.

--[[  -------------- Regular Grid Importer --------------------------------------------------
|
|  A Lua script to import Regular Grid data into Aspire
]]

require "strict"
--require("mobdebug").start()

g_title = "Regular Grid Importer"
g_version = "1.4"


--[[  -------------- ReadReliefFromRegularGrid --------------------------------------------------
|
|  Reads in a regular grid
|  Parameters:
|	options         -- The set of options including where to find the regular grid file
|   script_path     -- Where to locate the html pages
|     
|  Return Values:
|  A relief generated from the data and a z-scale to correct the height of the component
]]
function ReadReliefFromRegularGrid(options, script_path)

-- Unpack the plugin
local plugin = options.plugin
if not plugin then
	MessageBox("Unable to load file type")
	return nil
end

-- First pass of the file determine the size of the relief
-- Set up the bounds
local bounds = Box3D()

-- Keep track of pixel width and height
local pixel_width = 0
local pixel_height = 0

-- Use a table to keep track of how many pixels are in each row
-- We assume that each row has the same number of pixels
local pixel_widths = {}

-- Define a function to process the received points during the bounding box phase
local MergePoint = function(x, y, z)
	if x and y and z then
		-- Keep track of the points on a scanline
		local row_pixel_width = 0
		if not pixel_widths[y] then
			row_pixel_width = 1
			pixel_height = pixel_height + 1 -- we've found another row
		else
			row_pixel_width = pixel_widths[y] + 1
		end
		pixel_widths[y] = row_pixel_width
		if row_pixel_width > pixel_width then
			pixel_width = row_pixel_width
		end
		-- Update the bounds using the coordinates
		bounds:Merge(x, y, z)
	end
end

-- Set up a progress bar
local progress_bar = ProgressBar("Computing bounding box", ProgressBar.LINEAR)
local SetProgress = function(progress)
	progress_bar:SetPercentProgress(100 * progress + 0.5)
end

-- Warning functions
local warning_count = 0
local warning_lines = {}
local LogWarning = function(warning)
	-- Increment the warning count
    warning_count = 1 + warning_count
    -- Record the first 100 errors
    if warning_count <= 100 then
      warning_lines[warning_count] = warning
    end
end

-- Get the plugin to read the file
local read_result = plugin.Read(options.filename, MergePoint, SetProgress, LogWarning)
if not read_result then
	MessageBox("An error occurred while reading the file. Closing Gadget.")
	return nil
end

-- Attempt to free the ram, reliant on garbage collection
pixel_widths = {}

-- Report if there were any warnings in this file and
-- ask if the user wants to proceed
if warning_count > 0 then
  local warning_text = "The following problems were encountered with the file:\r\n\r\n"
  warning_text = warning_text ..table.concat(warning_lines, "\r\n")
  warning_text = warning_text .. "\r\n"
  if warning_count > 100 then
    warning_text = warning_text .. "and " .. tostring(warning_count - 100) .. " further warnings.\r\n"
  end
  warning_lines = {}
  local result = DisplayWarningDialog(options, script_path, warning_text)
  if result == 0 then
    return nil
  end
end

-- Might have an invalid bounding box
if bounds.IsInvalid then
  MessageBox("Failed to compute bounding box from the file. Closing Gadget.")
  return nil
end

-- Extract properties from the bounding box
local bounds_x_length = bounds.XLength
local bounds_y_length = bounds.YLength
local bounds_z_length = bounds.ZLength
local min_x = bounds.MinX
local min_y = bounds.MinY
local min_z = bounds.MinZ

-- Set up the dimensions of the relief initially from the bounding box
local rlf_x_length = bounds_x_length
local rlf_y_length = bounds_y_length
local rlf_z_length = bounds_z_length

-- Make a copy of the data_in_mm flag
local data_in_mm = options.data_in_mm

-- If the data doesn't have units then use the dpi setting to convert the 
-- dimensions into inches
if not plugin.HasUnits() then
    local thickness = options.job_thickness
	rlf_x_length = bounds_x_length / options.dpi
	rlf_y_length = bounds_y_length / options.dpi
	rlf_z_length = options.job_in_mm and 0.0393700787 * thickness or thickness
	data_in_mm = false
end

-- Scale the relief sizes into the appropriate units
if options.job_in_mm ~= data_in_mm  then
	local scale_factor = 1.0
	if options.job_in_mm then
		-- We're in mm and source data is in inches
		scale_factor = 25.4
	else
		-- We're in inches and the source data is in mm
		scale_factor = 0.0393700787
	end
	rlf_x_length = scale_factor * rlf_x_length
	rlf_y_length = scale_factor * rlf_y_length
	rlf_z_length = scale_factor * rlf_z_length
end

-- Render the model into the relief
-- Try and create the relief from this
local relief = Relief(pixel_width, pixel_height, rlf_x_length, rlf_y_length)
if not relief or relief.Error then
  MessageBox("Error - failed to create model")
  return nil
end

-- Create a function to set points in relief
local RenderPoint = function(x, y, z)
	if x and y and z then
		-- Work out the pixel co-ordinates
		z = z - min_z
		local pixel_x = 0.5 + (pixel_width  - 1) * (x - min_x) / bounds_x_length
		local pixel_y = 0.5 + (pixel_height - 1) * (y - min_y) / bounds_y_length
		local relief_z = relief:Get(pixel_x, pixel_y)
		-- Merge high
		if not IsTransparent(relief_z) then
			z = math.max(z, relief_z)
		end
		relief:Set(pixel_x, pixel_y, z)
	end
end

-- Update the progress bar
progress_bar:SetText("Rendering component")
progress_bar:SetPercentProgress(0)

-- Not interested in warnings during this pass so pass a no op warning function in
LogWarning = function(warning) end

-- Render the file into the relief
read_result = plugin.Read(options.filename, RenderPoint, SetProgress, LogWarning)

-- Finished with progress
progress_bar:Finished()

if not read_result then
	MessageBox("An error occurred while reading the file")
	return nil
end

-- Compute the z scale
local z_scale = rlf_z_length / bounds_z_length

return relief, z_scale
end

--[[  -------------- DisplayWarningDialog --------------------------------------------------
|
|  Displays the warnings dialog
|  Parameters:
|   options         -- The set of options for dialog
|   script_path     -- Where to locate the html pages
|   warning_text    -- The warning 
|     
|
|  Return Values:
|  1 on success,
|  0 if the user cancelled
|
]]
function DisplayWarningDialog(options, script_path, warning_text)

-- Prepare the dialog
local script_html = "file:" .. script_path .. "\\Regular_Grid_Warning.htm"
local dialog_name = options.gadget_title
local dialog = HTML_Dialog(false, script_html, 640, 480, dialog_name)

-- Set the gadget name and version
dialog:AddLabelField("GadgetTitle", options.gadget_title)
dialog:AddLabelField("GadgetVersion", options.gadget_version)

-- Set up the errors
dialog:AddTextField("WarningText", warning_text)

-- Show the dialog
if not dialog:ShowDialog() then
  return 0
end

return 1
end

--[[  -------------- DisplayDialog --------------------------------------------------
|
|  Displays the options dialog
|  Parameters:
|	options           -- The set of options for dialog
|   script_path       -- Where to locate the html pages
|   job            	  -- The job
|     
|
|  Return Values:
|  1 on success,
|  0 if the user cancelled
|  -1 on bad input
|
]]

OnFilePicker_ChooseFileButton = nil

function DisplayDialog(options, script_path, job)

-- Prepare the dialog
local script_html = "file:" .. script_path .. "\\Regular_Grid_Importer.htm"
local dialog_name = options.gadget_title
local dialog = HTML_Dialog(false, script_html, options.window_width, options.window_height, dialog_name)

-- Set the gadget name and version
dialog:AddLabelField("GadgetTitle", options.gadget_title)
dialog:AddLabelField("GadgetVersion", options.gadget_version)

-- Add a component name field
dialog:AddTextField("ComponentNameEdit", options.component_name)

-- Add file picker
dialog:AddLabelField("FileNameLabel", options.filename)
dialog:AddFilePicker(true, "ChooseFileButton", "FileNameLabel", false)

-- Add data in mm checkbox
dialog:AddCheckBox("DataInMMCheck", options.data_in_mm)

-- Validate the string input
local StringInputInvalid = function(input)
  return not input or input:len() == 0
end

-- Set up a button handler for the file choose button
-- Put it in the global table...
OnFilePicker_ChooseFileButton = function()
	-- Clear the selected plugin
	options.plugin = nil
	-- See if we can find a plugin for this filetype
	local filepath = dialog:GetLabelField("FileNameLabel")
	if StringInputInvalid(filepath) then
		return false
	end
	local pathutils = require("pathutils")
	local _, _, ext = pathutils.SplitPath(filepath)
	if not ext then
		MessageBox("File doesn't have an extension - unable to determine type!")
		return false
	end
	ext = ext:upper()
	for _, candidate in ipairs(options.plugins) do
		if candidate.SupportsExtension(ext) then
			options.plugin = candidate
			break
		end
	end
	if not options.plugin then
		dialog:UpdateLabelField("FileNameLabel", "")
		MessageBox("File type not supported!")
		return false
	end
	-- Update the component name based on the file
	local filename = pathutils.GetFileNameWithoutExtension(filepath)
	dialog:UpdateTextField("ComponentNameEdit", filename)
	-- Update the data in mm check
	local checkbox_html = [[<input type="checkbox" ID="DataInMMCheck">Data in mm</input>]]
	local has_units = options.plugin.HasUnits()
	local html = nil
	if has_units then
		html = [[<span id="DataInMMSpan">]] .. checkbox_html ..[[</span>]]
	else
		html = [[<span id="DataInMMSpan" style="visibility:hidden;">]] .. checkbox_html ..[[</span>]]
	end
	dialog:SetOuterHtml("DataInMMSpan", html)
	return true
end

-- Show the dialog
local dialog_result = dialog:ShowDialog()

-- Remove the function from the global table
OnFilePicker_ChooseFileButton = nil

-- Exit if the user cancelled
if not dialog_result then
  return 0
end

-- Update the options from this dialog
options.component_name = dialog:GetTextField("ComponentNameEdit")
options.filename = dialog:GetLabelField("FileNameLabel")
options.data_in_mm = dialog:GetCheckBox("DataInMMCheck")

-- Check the filename
if StringInputInvalid(options.filename) then
	MessageBox("No filename!")
	return -1
end

-- Check component name
if StringInputInvalid(options.component_name) then
	MessageBox("No component name!")
	return -1
end

-- Check there's a plugin selected
if not options.plugin then
	MessageBox("Unsupported file format!")
	return -1
end

return 1

end

--[[  -------------- LoadPlugins --------------------------------------------------
|
|  Load the plugins
|  Parameters:
|   script_path     -- Where the gadget is located
|
|  Return Values:
|	The list of loaded plugins
|
]]

function LoadPlugins(script_path)

-- Extend the package path
local old_package_path = package.path
package.path = package.path .. ";" .. script_path .. "\\Plugins\\?.luax"

-- Scan the directory for a list of lua files
local dir_reader = DirectoryReader()
dir_reader:BuildDirectoryList(script_path .. "\\Plugins", false)
dir_reader:GetFiles("*.luax", true, false)

-- Load path utils
local pathutils = require("pathutils")

local plugins = {}
local plugin_count = dir_reader:NumberOfFiles()
for i = 1, plugin_count do
	local file_entry = dir_reader:FileAtIndex(i)
	local filename = pathutils.GetFileNameWithoutExtension(file_entry.Name)
	local p = require(filename)
	if p then
		table.insert(plugins, p)
	end
end

-- Restore the package path
package.path = old_package_path

return plugins
end



--[[  ------------------------ main --------------------------------------------------
|
|  Entry point for script
|
]]
function main(script_path)

-- Check for Aspire
if not IsAspire() then
	MessageBox("This gadget requires Aspire!")
	return false
end

if GetBuildVersion() < 7.5 then
	MessageBox("This gadget requires Aspire 4.5 or later!")
	return false
end

-- Check for the existence of a job
local job = VectricJob()
if not job.Exists then
	MessageBox("No job open!")
	return false
end

-- Extend the package path for any common utilities
package.path = package.path .. ";" .. script_path .. "\\Common\\?.luax"

-- Get a material block for the material thickness
local material_block = MaterialBlock()

-- Set up our options
local options = {
  gadget_title = g_title,
  gadget_version = g_version,
  window_width = 640,
  window_height = 260,
  component_name = "Regular Grid Component",
  job_in_mm = job.InMM,
  job_thickness = material_block.Thickness,
  dpi = 300,
  filename = "",
  data_in_mm = false,
  plugin = nil
}

-- Load the plugins into options
options.plugins = LoadPlugins(script_path)

-- Check we have any plugins
if #options.plugins == 0 then
	MessageBox("No plugins found. Can't import any files!")
	return false
end

-- Display the dialog
-- Looping until either the user cancels or presses ok
local dialog_code = -1
while dialog_code == -1 do
  dialog_code = DisplayDialog(options, script_path, job)
end

-- Did the user cancel
if dialog_code == 0 then
  return false
end

-- Read in the point cloud file using the plugin
local relief, z_scale = ReadReliefFromRegularGrid(options, script_path)
if not relief then
  return false
end

-- Add the relief to the manager
local component_manager = job.ComponentManager
local id = component_manager:AddRelief(relief, CombineMode.Add, options.component_name)
local component = id and component_manager:FindObjectWithId(id) or nil
if component then
	component:SetZScale(z_scale)
end
component_manager:UpdateCompositeModel()

-- Required to return boolean
return true

end